import random
import numpy as np

def split_data(X,y,num_test, seed = 123):

    random.seed(seed)
    n,_ = X.shape

    test_idx  = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]

    return X_train,y_train, X_test, y_test

def split_data_vcv(X,y,vcv, num_test, seed = 123):
    
    random.seed(seed)
    n,_ = X.shape

    test_idx  = random.sample(range(n), k = num_test)
    train_idx = list( set(range(n)) - set(test_idx) )

    X_train, X_test = X[train_idx,:], X[test_idx,:]
    y_train, y_test = y[train_idx]  , y[test_idx]
    vcv_train, vcv_test = vcv[train_idx,:][:,train_idx], vcv[test_idx,:][:,test_idx]

    return X_train, X_test, y_train, y_test, vcv_train, vcv_test

def k_folds(X, folds = 3, seed = 123):
    """
    test_indx, train_indx
    """
    # X = X_train
    # folds = 4
    random.seed(seed)
    
    n,_ = X.shape
    all_index = list(range(n))
    random.shuffle(all_index)

    window = n/folds

    k_folds = []

    i = 0
    while True:

        init_indx = i
        end_indx  = i + window

        test_indx = all_index[round(init_indx):round(end_indx)]
        train_indx = list(set(all_index) - set(test_indx))
        # print(init_indx, end_indx)
        k_folds.append([test_indx, train_indx])

        i += window
        if i >= n:
            break

    # len(k_folds)
    return k_folds

def evaluate_folds(X, y, myFolds, model, tmp_params):

    # print(kwargs)
    # kwargs = {'c': 0.4, 'lambda': 0.1}
    # params = {'gamma': 0.4, 'lambda': 0.1}
    # params = tmp_params
    # model = phyloKRR(kernel='rbf')

    model.set_params(tmp_params)
    # model.get_params()

    all_errs = []
    for test_indx, train_indx in myFolds:
        # print(len(test_indx), len(train_indx))
        X_train,y_train = X[train_indx,:], y[train_indx]
        X_test,y_test = X[test_indx,:], y[test_indx]

        model.fit(X_train, y_train)

        # print(np.var(X_train))
        # print(np.var(model.X))
        # print(np.var(model.alpha))

        tmp_err = model.score(X_test, y_test, metric = 'rmse')
        all_errs.append(tmp_err)

    # return np.mean(all_errs)
    return np.median(all_errs)

def k_fold_cv(X, y, vcv, model, num_folds):
    """
    k-fold cross-validation with covariance matrix
    """
    n, p = X.shape
    fold_size = n // num_folds
    mse_sum = 0

    for i in range(num_folds):

        test_idx = list(range(i * fold_size, (i + 1) * fold_size))
        train_idx = list(set(range(n)) - set(test_idx))

        X_train, X_test = X[train_idx, :], X[test_idx, :]
        y_train, y_test = y[train_idx], y[test_idx]

        vcv_train = vcv[train_idx,:][:,train_idx]
        vcv_test = vcv[test_idx,:][:,test_idx]

        model.fit(X_train, y_train, vcv = vcv_train)
        mse_sum += model.score(X_test, y_test, vcv_test)

    return mse_sum / num_folds

def k_fold_cv_random(X, y, vcv,
                     model, 
                     params,
                     folds = 3, 
                     sample = 500,
                     verbose = True,
                     seed = 123
                     ):
    """
    Random search for hyperparameter tuning using k-fold cross-validation
    and covariance matrix
    """
    
    np.random.seed(seed=seed)
    # make random choice from the grid of hyperparameters
    all_params = params.keys()
    tested_params = np.zeros((sample, len(all_params)))
    for n,k in enumerate(all_params):
        tested_params[:,n] = np.random.choice(params[k], sample)

    if verbose:
        # check tested_params are unique
        tested_params = np.unique(tested_params, axis=0)
        print("Number of unique hyperparameters: ", tested_params.shape[0])
    
    all_errors = []
    for vec in tested_params:
        tmp_params = dict(zip(all_params, vec))
        model.set_params(**tmp_params)
        tmp_err = k_fold_cv(X, y, vcv, model, folds)
        all_errors.append([tmp_params, tmp_err])

    best_ = sorted(all_errors, key=lambda kv: kv[1], reverse=False)[0]

    if verbose:
        print("CV score: ", best_[1])

    return best_[0]

def PGLS(X, y, vcv):
    """
    Generalized Least Squares with phylogenetic covariance matrix
    as the weight matrix
    """
    n,p = X.shape
    Oinv = np.linalg.inv(vcv)
    X_ = np.hstack((np.ones((n,1)), X))
    beta = np.linalg.inv(X_.T @ Oinv @ X_) @ X_.T @ Oinv @ y
    return beta




import numpy as np

from phylokrr.utils import split_data, k_fold_cv_random
from phylokrr.kernels import KRR


class PhylogeneticRegressor:

    def __init__(self, X, y, cov, kernel='rbf') -> None:

        self.X = X
        self.y = y
        self.vcv = self.read_cov(cov)
        self.model = KRR(kernel=kernel)

        self.check_shapes()

        # weighting data by the cov matrix
        self.P, self.X1, self.y1 = self.P_mat(chol = False)


        self.hyperparamter_space = {}
        self.P_inv = np.array([])
        self.test_size = None

    
    def check_shapes(self):
        assert self.vcv.shape[0] == self.vcv.shape[1], 'not square cov matrix'
        assert self.X.shape[0] == self.y.shape[0] == self.vcv.shape[0], 'dimensions do not match'

        # assert self.check_matrix(self.vcv), 'not symmetric matrix'
    def check_symmetric(self, a, tol=1e-8):
        return np.all(np.abs(a-a.T) < tol)        

    def read_cov(self, cov):

        if isinstance(cov, np.ndarray):
            return cov
        
        else:
            return np.loadtxt(cov, delimiter=',')
        
    def P_mat(self, chol = False):

        if chol:
            P = np.linalg.cholesky( np.linalg.inv( self.vcv ) )

        else:
            Oinv = np.linalg.inv( self.vcv )
            L,Q  = np.linalg.eig( Oinv )
            P  = Q @ np.diag( np.sqrt( 1/L ) ) @ Q.T

        return P, P @ self.X, P @ self.y
    
    def split_data(self, seed = 123):
        """
        split weighted data
        """
        num_test = round( self.X.shape[0] * self.test_size )

        return split_data(self.X1, self.y1, num_test, seed = seed)
    
    def set_hyperparameter_space(self, params):
        self.hyperparamter_space = params

    def fit(self, 
            cv = 3, 
            sample = 100, 
            test_size = 0.4, 
            seed = 123,
            verbose = True
            ):
        
        """
        Train model with weighted data

        Parameters
        ----------
        cv : int, default = 3
            number of cross validation rounds 
            for hyperparameter tuning with the 
            training set
            
        sample: int, default = 100
            sample from the hyperparameter space 
            for hyperparameter tuning with the
            training set
        """
        self.test_size = test_size

        X_train, y_train, X_test, y_test = self.split_data(seed = seed)

        assert len(self.hyperparamter_space), 'set hyperparameter space'

        # X_train.shape
        hyperparameters = k_fold_cv_random(
                            X_train,
                            y_train,
                            self.model,
                            self.hyperparamter_space,
                            folds  = cv,
                            sample = sample,
                            verbose=verbose
                            )
        
        if verbose:
            print('hyperparameters: %s' % hyperparameters)

        self.model.set_params(**hyperparameters)
        self.model.fit(X_train, y_train)

        if verbose:
            testing_error = self.model.score(X_test, y_test, metric='rmse')
            training_error = self.model.score(X_train, y_train, metric='rmse')
            print('Training error: %s' % training_error)
            print('Testing error: %s' % testing_error)
            
    def set_params(self, hyperparameters, fit = False, test_size = 0.4, seed = 123):
        """
        Set new hyperparameters

        Parameters
        ----------
        refit : bool, default = False
            refit model after setting hyperparamters
            
        test_size: float, default = 0.4
            if refit, the test size proportion
        """
        self.model.set_params(**hyperparameters)

        if fit:
            self.test_size = test_size

            X_train, y_train, X_test, y_test = self.split_data(seed = seed)
            self.model.fit(X_train, y_train)

            testing_error  = self.model.score(X_test, y_test, metric='rmse')
            training_error = self.model.score(X_train, y_train, metric='rmse')

            print('Training error: %s' % training_error)
            print('Testing error: %s' % testing_error)


    def sample_feature(self, feature, quantiles, sample, integer = False):
        """
        sample unweighted column
        """
        # X = X0
        # feature =0
        Xpdp = self.X[:,feature]

        if integer:
            return np.sort(np.unique(Xpdp))
        
        q = np.linspace(
            start = quantiles[0],
            stop  = quantiles[1],
            num   = sample
        )

        # sample from the original feature space
        Xq = np.quantile(Xpdp, q = q)

        return Xq    

    def pdp(self, feature, integer = False, quantiles = [0,1], sample = 70):
        """
        partial depedence plot values.

        Parameters
        ----------
        feature : int

        Returns
        -------
        PDP of the transformed data
        """
        if not len(self.P_inv):
            self.P_inv = np.linalg.inv(self.P)

        Xq = self.sample_feature(feature, quantiles, sample, integer)
        
        pdp_values = []
        for n in Xq:
            # copy original values
            X_tmp = self.X.copy()
            # make original values with 
            # modified feature column
            X_tmp[:,feature] = n
            # weight the whole dataset as
            # the model learned from the 
            # weighted data
            X_tmp_test = self.P @ X_tmp

            pdp_values.append(
                np.mean(
                    # make rows unweighted
                    # (i.e., again 'correlated')
                    self.P_inv @ self.model.predict(X_tmp_test)
                )
            )
 
        out = np.hstack((
            Xq.reshape(-1,1), 
            np.array(pdp_values).reshape(-1,1)
        ))
        return out
    
    def FeatureImportance(self, seed = 123):
        """
        permutation feature importance
        """        
        assert len(self.model.alpha), 'model needs to be fitted'

        np.random.seed(seed=seed)
        

        X_train, y_train, X_test, y_test = self.split_data(seed = seed)
        self.model.fit(X_train, y_train)
        error_orig = self.model.score(X_test, y_test)


        out = []
        for i in range(X_test.shape[1]):
            # Create a copy of X_test
            X_test_copy = X_test.copy()

            # Scramble the values of the given predictor
            X_test_copy[:,i] = np.random.permutation(X_test_copy[:,i])
                        
            # Calculate the new RMSE
            error_perm = self.model.score(X_test_copy, y_test)

            out.append(error_perm - error_orig)

        return out
    
class PhylogeneticLogisticRegressor:
    def __init__(self) -> None:
        pass





import numpy as np

def R2(y, r):
    """
    Coefficient of determination
    r = y - y_pred 
    """
    # r = y - y_pred

    u = r.T @ r
    v = y.T @ y - len(y) * np.mean(y)**2
    return 1 - (u/v)

def rmse(r):
    """
    Root Mean Squared Error
    r = y - y_pred
    """
    # r = y - y_pred
    return np.sqrt( np.mean( r**2 ) )



from phylokrr.kernels import KRR

class PagelsLambda:

    def __init__(self, bounds = (0.1, 1), model = KRR(kernel='rbf')):
        self.bounds = bounds
        self.model = model

    def fit(self, model, X, y, vcv):
        pass





import numpy as np
from phylokrr.metrics import rmse, R2

def distance_matrix(a, b):
    """
    l2 norm squared matrix
    """
    return np.linalg.norm(a[:, None, :] - b[None, :, :], axis=-1)**2

def RBF_kernel(a, b, gamma):
    """
    Radial Basis Function
    """        
    tmp_rbf = -gamma * distance_matrix(a, b)
    np.exp(tmp_rbf, tmp_rbf) # RBF kernel. Inplace exponentiation
    return tmp_rbf

def linear_kernel(a, b, c):
    """
    Linear Kernel
    """
    XXt = a.dot(b.T)
    C = c * np.ones(XXt.shape)

    return XXt + C



class KRR:

    def __init__(self, kernel = 'rbf', fit_intercept = True, check_cov = False) -> None:
        """
        Weighted Kernel Ridge Regression

        kernel: str, default = 'rbf'
            kernel type

        fit_intercept: bool, default = True
            fit intercept term

        check_cov: bool, default = False
            Check covariance matrix, i.e., P.T @ P == vcv^-1
        """

        self.kernel = kernel

        if self.kernel == 'rbf':
            self.params = {'gamma': 0.1, 'lambda': 0.1}

        else:
            self.params = {'c': 0.1, 'lambda': 0.1}

        self.fit_intercept = fit_intercept
        self.check_cov = check_cov
        
        # internal
        self.intercept = 0
        self.alpha = np.array([])
        self.X = np.array([])

        self.chol = False

    def set_params(self, **params):

        if self.kernel == 'rbf':
            self.params['gamma'] = params['gamma']

        else:
            self.params['c'] = params['c']

        self.params['lambda'] = params['lambda']

    def get_params(self):
        return self.params
    
    def P_mat(self, vcv):
        """
        get the square root of the inverse of the
        covariance matrix.

        vcv: np.array, default = None
            covariance matrix
        """
        
        if isinstance(vcv, type(None)):
            return None
        
        if self.check_cov:
            self.assert_COV_sym(vcv)

        if self.chol:
            C = np.linalg.cholesky( vcv )
            P = np.linalg.inv( C )

        else:
            L,Q  = np.linalg.eig( vcv )
            P  = Q @ np.diag( L**(-1/2) ) @ Q.T 

        if self.check_cov:
            self.assert_COV_decom(P, vcv)

        return P
    
    def assert_COV_sym(self, vcv, tol=1e-8):
        assert np.all(np.abs(vcv-vcv.T) < tol), 'not symmetric matrix'

    def assert_COV_decom(self, P, vcv):
        """
        check if P.T @ P == vcv^-1
        expensive calculation for large matrices
        """
        assert np.all(np.round(P.T @ P, 2) == np.round(np.linalg.inv(vcv), 2)), "P.T @ P != vcv^-1"

    def fit(self, X, y, vcv = None):
        """
        Fit the model

        vcv: np.array, default = None
            covariance matrix

        """
        # X = X_train
        # y = y_train
        # vcv = vcv_train
        
        self.X = X
        # self.y = y
        P = self.P_mat(vcv)

        if self.kernel == 'rbf':
            K_train = RBF_kernel(self.X, self.X, self.params['gamma'])

        else:
            K_train = linear_kernel(self.X, self.X, self.params['c'])

        self.alpha = self.opt_alpha(K_train, y, self.params['lambda'], P)
        
        # intercept term 
        if self.fit_intercept:
            self.intercept = np.mean(y - K_train @ self.alpha)

    def predict(self, X_test):

        assert len(self.alpha) > 0, "The model needs to be fitted first"

        if self.kernel == 'rbf':
            K_test = RBF_kernel(X_test, self.X, self.params['gamma'])

        else:
            K_test = linear_kernel(X_test, self.X, self.params['c'])
        
        return K_test @ self.alpha + self.intercept
    
    def score(self, X_test, y_test, vcv_test, metric = 'rmse'):
        """
        
        Calculates score considerint the covariance matrix
        This error is the same as the in objective function

        X_test: np.array
            testing data
        
        y_test: np.array
            testing targets
        
        vcv_test: np.array
            testing covariance matrix. If None, then it
            will return the standard RMSE and R^2

        metric: str, default = 'rmse'
            weighted root mean squared error (the same 
            that was used to construct the objective function).
            if metric = 'r2', then it returns the R^2 which
            also considers the covariance matrix, see Willett & Singer (1988)

        Returns:
        --------
        float
            score

        Reference:
        ---------
        Willett, J. B., & Singer, J. D. (1988). 
        Another cautionary note about R 2: Its use 
        in weighted least-squares regression analysis. 
        The American Statistician, 42(3), 236-238.
        """

        y_pred = self.predict(X_test)
        r = y_pred - y_test
        
        if isinstance(vcv_test, type(None)):
            P = np.eye(X_test.shape[0])

        else:
            P = self.P_mat(vcv_test)

        werr = P @ r # weighted residuals
        Py   = P @ y_test # weighted targets

        if metric == 'rmse':
            return rmse(werr)
        
        else:
            return R2(Py, werr)
            
        # if metric == 'rmse':
        #     return np.sqrt( np.mean( r**2 ) )
        
        # else:
        #     u = r.T @ r
        #     v = y_test.T @ y_test - n * np.mean(y_test)**2
        #     return 1 - (u/v)
    
    def opt_alpha(self, K, y, reg_lam = None, P = None):
        # Y = y
        # X = X
        # m = None
        n,_ = self.X.shape
        
        I = np.eye(K.shape[0])
        nlI = n * reg_lam * I

        if isinstance(P, type(None)):
            return np.linalg.solve(K + nlI, y)
        
        else:
            return P @ np.linalg.solve(P @ K @ P + nlI,  P @ y)
            # return np.linalg.solve(P.T @ P @ K + nlI,  P.T @ P @ y)


class LKRR:

    def __init__(self, kernel = 'rbf') -> None:

        self.kernel = kernel

        if self.kernel == 'rbf':
            self.params = {'gamma': 0.1, 'lambda': 0.1}

        else:
            self.params = {'c': 0.1, 'lambda': 0.1}

        # internal
        self.alpha = []
        self.X = []
        self.y = []

    def p1(self, X, beta):
        return 1/(1 + np.exp(X.dot(beta)))

    def p0(self, X, beta):
        EX = np.exp(X.dot(beta))
        return EX/(1 + EX)

    def set_params(self, params):
        self.params = params

    def get_params(self):
        return self.params
    
    def fit(self, X, y):
        
        self.X = X
        self.y = y

        if self.kernel == 'rbf':
            K_train = RBF_kernel(self.X, self.X, self.params['gamma'])

        else:
            K_train = linear_kernel(self.X, self.X, self.params['c'])

        self.alpha = self.opt_alpha(K_train, self.y, self.params['lambda'])

    def predict(self, X_test):

        assert len(self.alpha) > 0, "The model needs to be fitted first"

        if self.kernel == 'rbf':
            K_test = RBF_kernel(X_test, self.X, self.params['gamma'])

        else:
            K_test = linear_kernel(X_test, self.X, self.params['c'])
        
        return K_test @ self.alpha
    
    def score(self, X_test, y_test, metric = 'rmse'):

        y_pred = self.predict(X_test)

        if metric == 'rmse':
            return np.sqrt( np.mean( (y_pred - y_test)**2 ) )
    
        else:
            # r2
            u = ((y_test - y_pred)**2).sum()
            v = ((y_test - y_test.mean())** 2).sum()

            return 1 - (u/v)
    
    def rmse(self, K, alpha, Y):

        return np.sqrt( np.mean( (K.dot(alpha) - Y)**2 ) )
    
    def opt_alpha(self, K, y, reg_lam = None):
        # Y = y
        # X = X
        # m = None
        K_Idx = K + reg_lam * np.diag( np.ones( K.shape[0] ) )
        return np.linalg.inv( K_Idx ).dot( y )


# def P_mat(vcv, chol = False, corr = False):
    
#     if corr:
#         Kr = np.diag(1/np.sqrt(np.diag(vcv)))
#         vcv = Kr @ vcv @ Kr

#     if chol:
#         P = np.linalg.cholesky( np.linalg.inv( vcv ) )

#     else:
#         Oinv = np.linalg.inv( vcv )
#         L,Q  = np.linalg.eig( Oinv )
#         P  = Q @ np.diag( np.sqrt( 1/L ) ) @ Q.T

#     return P


# ## logistic data
# data = np.loadtxt('../data/test_log_data.csv', delimiter=',')

# # get covariance matrix
# cov = np.loadtxt('../data/test_cov.csv',delimiter=',')


# [n,p] = np.shape(data)
# num_train = int(0.5*n)

# X,y = data[:,:-1], data[:,-1]
# P = P_mat(cov,corr=True)

# X = P @ X
# y = P @ y





# sample_train = X[0:num_train,:]
# sample_test  = X[num_train: ,:]
    
# label_train = y[0:num_train]
# label_test  = y[num_train: ]
    


# # import matplotlib.pyplot as plt
# # plt.scatter(X[:,0], X[:,2], c = y, alpha=0.6)

# import warnings

# def p1(X, beta):
#     return 1/(1 + np.exp(X.dot(beta)))

# def p0(X, beta):
#     EX = np.exp(X.dot(beta))
#     return EX/(1 + EX)

# def error_est(X, y,  beta, norm = False):
#     if norm:
#         return np.linalg.norm(p1(X,beta) - y, ord = 2)
#         # pass
#     else:
#         y_pred = (p1(X, beta) >= 0.5).astype(float)
#         return np.mean(y_pred != y)

# def NewtonRaphsonStep(X, y, beta):

#     Lp = -X.T @ ( y - p1(X, beta) )
#     W  = np.diag( p1(X, beta) * p0(X, beta) )

#     Lpp = -X.T @ W @ X
#     return  beta - np.linalg.inv(Lpp) @ Lp


# n_updts = 20
# beta0 = np.array(np.random.normal(size = p - 1))
# beta = np.copy(beta0)

# epsilon = 1e-6
# for i in range(n_updts):
#     # i = 0
#     beta_old = np.copy(beta)

#     with warnings.catch_warnings():
#         warnings.simplefilter("ignore")

#         beta = NewtonRaphsonStep(sample_train, label_train, beta)

#         if np.any( np.isnan(beta) ):
#             print('beta re-redifined')
#             beta = np.random.normal(size = p - 1)

#         error = error_est(sample_test, label_test, beta, norm=True)
#         print("update = %i, error = %s" %(i, round(error, 5))) 

#     if error <= epsilon:
#         print('error convergence')
#         break

#     beta_norm = np.linalg.norm(beta_old - beta, ord = np.inf )
#     if np.any( beta_norm <= epsilon  ):
#         print('coefficient convergence')
#         break

# y_pred = (p1(sample_test, beta) >= 0.5).astype(float)

# from sklearn.metrics import confusion_matrix
# confusion_matrix(label_test,y_pred)



#!/usr/bin/env python

import setuptools
from distutils.core import setup

dependencies = [
    "numpy"
]

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(name = "phylokrr",
      version = '0.4.2',
      long_description = readme,
      long_description_content_type = 'text/markdown',
      packages = ['phylokrr'],
      package_dir = {'phylokrr': 'src'},
      python_requires='>=3.5',
      install_requires = dependencies,
      zip_safe = False,
      classifiers = [
          'Programming Language :: Python :: 3',
      ]
    )


